#include <pcap.h>
#include <signal.h>

#define DEFAULT_SNAPLEN 128
pcap_t  *pd;
int verbose = 1;
struct pcap_stat pcapStats;

/* *************************************** */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <errno.h>
#include <sys/poll.h>
#include <time.h>
#include <netinet/ip.h>
#include <net/ethernet.h>     /* the L2 protocols */


#define PAGE_SIZE    4096
#define BUCKET_LEN    128
#define NUM_SLOTS     512

#define DEFAULT_DEVICE "eth1" /* "e1000" */

static u_int numSlots = NUM_SLOTS;
unsigned long numPkts = 0, numBytes = 0;

typedef struct flowSlotInfo {
  u_short totSlots;
  u_int32_t totPkts, totLost;
  u_short insertIdx;
  u_short removeIdx;
} FlowSlotInfo;

typedef struct flowSlot {
  u_char     slotState; /* 0=empty, 1=full */
  u_short    slotLen;
  u_char     bucket[BUCKET_LEN];
} FlowSlot;

static int memSlotsLen, fd;
static unsigned char *buf;
static FlowSlot *slots;
static FlowSlotInfo *slotsInfo;
static u_int32_t totPkts, totLost;
static struct timeval startTime;

int numPolls=0;

int pcap_stats(pcap_t *p, struct pcap_stat *ps) {
  ps->ps_recv = slotsInfo->totPkts-pcapStats.ps_recv,
    ps->ps_drop = slotsInfo->totLost-pcapStats.ps_drop;
  return(1); /* OK */
}

char * pcap_lookupdev(char *errbuf) {
  return(DEFAULT_DEVICE);
}

pcap_t * pcap_open_live(char *device, int snaplen, int promisc, int to_ms, char *ebuf) {
  char procName[64];

  snprintf(procName, sizeof(procName), "/proc/net/%s-ring", device);

  memSlotsLen = numSlots*sizeof(FlowSlot)+sizeof(FlowSlotInfo);
  /* round it to PAGE_SIZE (4096) boundary */
  memSlotsLen = ((memSlotsLen/PAGE_SIZE)+1)*PAGE_SIZE;

  fd = open(procName, O_RDWR /* O_RDONLY|O_LARGEFILE */);
  if(fd <= 0){
    printf("Unable to open '%s' entry [%s]\n", procName, strerror(errno));
    return(NULL);
  } else
    printf("Reading from %s\n", procName);

  buf = (char *)mmap(NULL, memSlotsLen, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
  if(buf == MAP_FAILED) {
    perror("mmap() failed");
    return (NULL);
  } else {
    printf("mmap() returned 0x%x\n", buf);
  }

  slotsInfo = (FlowSlotInfo *)buf;
  slots     = (FlowSlot *)&buf[sizeof(FlowSlotInfo)];

  return((pcap_t*)slots);
}

int pcap_loop(pcap_t *p, int cnt, pcap_handler callback, u_char *user) {
  int polling = 0;

  pcapStats.ps_recv = slotsInfo->totPkts, pcapStats.ps_drop = slotsInfo->totLost;
  
  while(1) {
    if(slots[slotsInfo->removeIdx].slotState == 1) {

      callback(DEFAULT_DEVICE, 
	       (const struct pcap_pkthdr*)slots[slotsInfo->removeIdx].bucket,
	       (const u_char*)&slots[slotsInfo->removeIdx].bucket[sizeof(struct pcap_pkthdr)]);
      slots[slotsInfo->removeIdx].slotState = 0;
      slotsInfo->removeIdx = (slotsInfo->removeIdx+1) % NUM_SLOTS;
    } else if(!polling) {
      struct pollfd pfd;

      /* Sleep when nothings happening */
      pfd.fd = fd;
      pfd.events = POLLIN|POLLERR;
      pfd.revents = 0;
      poll(&pfd, 1, -1);
      numPolls++;
    }
  } /* while() */

  return(0);
}

void pcap_close(pcap_t *p) {
  munmap(buf, memSlotsLen);
  close(fd);
}

/* *************************************** */
/*
 * The time difference in microseconds
 */
long delta_time (struct timeval * now,
                 struct timeval * before) {
  time_t delta_seconds;
  time_t delta_microseconds;
                                                                                                                                           
  /*
   * compute delta in second, 1/10's and 1/1000's second units
   */
  delta_seconds      = now -> tv_sec  - before -> tv_sec;
  delta_microseconds = now -> tv_usec - before -> tv_usec;
                                                                                                                                           
  if(delta_microseconds < 0) {
    /* manually carry a one from the seconds field */
    delta_microseconds += 1000000;  /* 1e6 */
    -- delta_seconds;
  }                                                                                                                                           
  return((delta_seconds * 1000000) + delta_microseconds);
}

/* ******************************** */

void sigproc(int sig) {
  struct pcap_stat pcapStat;
  struct timeval endTime;
  float deltaSec;

  gettimeofday(&endTime, NULL);

  deltaSec = (float)delta_time(&endTime, &startTime)/1000000;

  if(pcap_stats(pd, &pcapStat) >= 0)
    fprintf(stderr, "pcap_stats: %u pkts rcvd/%u pkts dropped [tot=%d][%.1f %%]\n",
	    pcapStat.ps_recv, pcapStat.ps_drop, pcapStat.ps_recv-pcapStat.ps_drop,
	    pcapStat.ps_recv == 0 ? 0 : (float)(pcapStat.ps_drop*100)/(float)pcapStat.ps_recv);

  fprintf(stderr, "%u pkts [%.1f pkt/sec]\n%u bytes [%.1f Mbit]\n", 
	  numPkts, (float)numPkts/deltaSec,
	  numBytes, (float)8*numBytes/(deltaSec*1000000));

  exit(0);
}

/* ****************************************************** */

static char hex[] = "0123456789ABCDEF";

char* etheraddr_string(const u_char *ep, char *buf) {
  u_int i, j;
  char *cp;

  cp = buf;
  if ((j = *ep >> 4) != 0)
    *cp++ = hex[j];
  else
    *cp++ = '0';

  *cp++ = hex[*ep++ & 0xf];

  for(i = 5; (int)--i >= 0;) {
    *cp++ = ':';
    if ((j = *ep >> 4) != 0)
      *cp++ = hex[j];
    else
      *cp++ = '0';

    *cp++ = hex[*ep++ & 0xf];
  }

  *cp = '\0';
  return (buf);
}

/*
 * A faster replacement for inet_ntoa().
 */
char* _intoa(unsigned int addr, char* buf, u_short bufLen) {
  char *cp, *retStr;
  u_int byte;
  int n;

  cp = &buf[bufLen];
  *--cp = '\0';

  n = 4;
  do {
    byte = addr & 0xff;
    *--cp = byte % 10 + '0';
    byte /= 10;
    if (byte > 0) {
      *--cp = byte % 10 + '0';
      byte /= 10;
      if (byte > 0)
	*--cp = byte + '0';
    }
    *--cp = '.';
    addr >>= 8;
  } while (--n > 0);

  /* Convert the string to lowercase */
  retStr = (char*)(cp+1);

  return(retStr);
}

/* ************************************ */

char* intoa(unsigned int addr) {
  static char buf[sizeof "ff:ff:ff:ff:ff:ff:255.255.255.255"];

  return(_intoa(addr, buf, sizeof(buf)));
}

/* ****************************************************** */

char* proto2str(u_short proto) {
  static char protoName[8];

  switch(proto) {
  case IPPROTO_TCP:  return("TCP");
  case IPPROTO_UDP:  return("UDP");
  case IPPROTO_ICMP: return("ICMP");
  default:
    snprintf(protoName, sizeof(protoName), "%d", proto);
    return(protoName);
  }
}

/* ****************************************************** */

static int32_t thiszone;

void dummyProcesssPacket(u_char *_deviceId,
			 const struct pcap_pkthdr *h,
			 const u_char *p) {
  if(verbose) {
    struct ether_header ehdr;
    u_short eth_type;
    char buf1[32], buf2[32];
    struct ip ip;
    int s = (h->ts.tv_sec + thiszone) % 86400;

    printf("%02d:%02d:%02d.%06u ",
	   s / 3600, (s % 3600) / 60, s % 60,
	   (unsigned)h->ts.tv_usec);

    memcpy(&ehdr, p, sizeof(struct ether_header));
    eth_type = ntohs(ehdr.ether_type);
    printf("[%s -> %s] ",
	   etheraddr_string(ehdr.ether_shost, buf1),
	   etheraddr_string(ehdr.ether_dhost, buf2));

    if(eth_type == 0x0800) {
      memcpy(&ip, p+sizeof(ehdr), sizeof(struct ip));    
      printf("[%s ", intoa(ntohl(ip.ip_src.s_addr)));
      printf("-> %s] ", intoa(ntohl(ip.ip_dst.s_addr)));
    }

    printf("[caplen=%d][len=%d]\n", h->caplen, h->len);
  }
  
  if(numPkts == 0) gettimeofday(&startTime, NULL);
  numPkts++, numBytes += h->len;
}

/* *************************************** */

int32_t gmt2local(time_t t) {
  int dt, dir;
  struct tm *gmt, *loc;
  struct tm sgmt;

  if (t == 0)
    t = time(NULL);
  gmt = &sgmt;
  *gmt = *gmtime(&t);
  loc = localtime(&t);
  dt = (loc->tm_hour - gmt->tm_hour) * 60 * 60 +
        (loc->tm_min - gmt->tm_min) * 60;

  /*
   * If the year or julian day is different, we span 00:00 GMT
   * and must add or subtract a day. Check the year first to
   * avoid problems when the julian day wraps.
   */
  dir = loc->tm_year - gmt->tm_year;
  if (dir == 0)
    dir = loc->tm_yday - gmt->tm_yday;
  dt += dir * 24 * 60 * 60;

  return (dt);
}

/* *************************************** */

int main(int argc, char* argv[]) {
  char *device = NULL;
  char errbuf[PCAP_ERRBUF_SIZE];

  thiszone = gmt2local(0);

  if(argc == 2)
    device = argv[1];

  if(device == NULL) {
    if((device = pcap_lookupdev(errbuf)) == NULL) {
      printf("pcap_lookup: %s", errbuf);
      return(-1);
    }
  }
  printf("Capturing from %s\n", device);

  /* hardcode: promisc=0, to_ms=500 */
  if((pd = pcap_open_live(device, DEFAULT_SNAPLEN, 0, 500, errbuf)) == NULL) {
    printf("pcap_open_live: %s", errbuf);
    return(-1);
  }

  signal(SIGINT, sigproc);
  pcap_loop(pd, -1, dummyProcesssPacket, NULL);

  pcap_close(pd);
  return(0);
}


